use utf8;
use strict;
use warnings FATAL => 'all';

package Xyzzy::Document::XHTML;

use Xyzzy::Document -self;

field forcepretty => undef;

our %structuralnodes;
@structuralnodes{qw(html head body div table tr ul ol dl form select)} = ();

sub killnodes {
	foreach my $node (@_) {
		my $type = ref $node;
		if($type eq 'XML::LibXML::Text') {
			my $parent = $node->parentNode;
			if(defined $parent) {
				if(exists $structuralnodes{$parent->localName}) {
					my $text = $node->data;
					$parent->removeChild($node)
						if $text =~ /^\s*$/
				}
			}
#		} elsif($type eq 'XML::LibXML::Comment') {
#			$node->parentNode->removeChild($node)
		} else {
			killnodes($node->childNodes)
		}
	}
}

sub getChunks {
	my ($file, $cachethis) = @_;
	our %templatecache;

	if(exists $templatecache{$file}) {
		return $templatecache{$file};
	} else {
#		print STDERR '['.localtime, '] [notice] ', ref $self, " parsing $file\n";
		my $parser = $self->xmlParser;
		my $newdoc = $parser->parse_file($file);

		my @nodes = grep { $_->hasAttributes } $newdoc->documentElement->childNodes;
		killnodes(@nodes)
			if $cachethis || $self->forcepretty;
		$templatecache{$file} = \@nodes
			if $cachethis;
		return \@nodes;
	}
}

sub getChunk {
	my ($file, $id) = @_;
	our %fragmentcache;

	my $fragment;
	if(exists $fragmentcache{$file}) {
		$fragment = $fragmentcache{$file}{$id};
	} else {
		my $nodes = $self->getChunks($file, 1);
		my %fragments = map {
				my $i = $_->getAttribute('id');
				$i ? ($i, $_) : ()
			} @$nodes;
		$fragmentcache{$file} = \%fragments;
		$fragment = $fragments{$id};
	}
	die "Fragment '$id' not in '$file'\n"
		unless defined $fragment;

	$fragment = $fragment->cloneNode(1);
	$self->xml->adoptNode($fragment);

#	our $expr ||= new XML::LibXML::XPathExpression('.//*[@id]');
	my $expr = './/*[@id]';
	my %ids;
	foreach my $node ($fragment->findnodes($expr)) {
		$ids{$node->getAttribute('id')} = $node;
		$node->removeAttribute('id');
	}
	return ($fragment, \%ids);
}

sub addChunks {
	my $cachethis = shift;

	foreach my $file (@_) {
		my $nodes = $self->getChunks($file, $cachethis);
		
		foreach my $node (@$nodes) {
			my $id = $node->getAttribute('id');
			my $victim = $self->getElementById($id);
			next unless defined $victim;
			my $newnode = $node->cloneNode(1);
			$self->xml->adoptNode($newnode);
			$victim->replaceNode($newnode);
		}
	}
}

sub getFragment {
	my $dir = $self->cfg->fragmentdir;
	my $file = shift;
	return $self->getChunk("$dir/$file.xml", @_);
}

sub addTemplates {
	my $dir = $self->cfg->templatedir;
	$self->addChunks(1, map { "$dir/$_.xml" } @_);
}

sub addContent {
	my $dir = $self->cfg->contentdir;
	$self->addChunks(0, map { "$dir/$_.xml" } @_);
}

sub build {
	super;

	my $dtd = $self->createInternalSubset('html',
		'-//W3C//DTD XHTML 1.1//EN',
		'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd');

	my $html = $self->createElementNS('http://www.w3.org/1999/xhtml', 'html');
	$html->setAttribute('xml:lang', 'en');
	$self->setDocumentElement($html);

	my $head = $self->createElement('head');
	$head->setAttribute('id', 'head');
	$html->appendChild($head);

	my $body = $self->createElement('body');
	$body->setAttribute('id', 'body');
	$html->appendChild($body);
}

sub serialize {
	my $res = new Xyzzy::Response(mimetype => 'application/xml');

	if($self->req->msie) {
		$res->mimetype('text/html', charset => 'UTF-8');
		$res->content($self->toStringHTML);
	} else {
		$res->mimetype('application/xhtml+xml');
		$res->content($self->toString);
	}

	return $res;
}

sub finalize {
	foreach($self->findnodes('/*|/*/*|/*/head//*')) {
		$_->removeAttribute('id')
			if $_->hasAttributes
	}
	return super;
}
