use utf8;
use strict;
use warnings FATAL => 'all';

$0 = "xyzzy (initializing)";

use Time::HiRes qw(clock_gettime CLOCK_PROCESS_CPUTIME_ID CLOCK_MONOTONIC);

my $wall_proc_start = clock_gettime(CLOCK_MONOTONIC)
	- clock_gettime(CLOCK_PROCESS_CPUTIME_ID);

use Encode qw(is_utf8 encode_utf8);
use CGI::Fast;
use Sys::Syslog qw(openlog closelog syslog);
use IO::Handle;

use Xyzzy;
use Xyzzy::Header;
use Xyzzy::Status;
use Xyzzy::Context;

sub concat {
	return undef if grep !defined, @_;
	return join('/', @_);
}

my $file = $ARGV[0];
my $class = 'Xyzzy';

openlog('xyzzy', 'nofatal,pid', Sys::Syslog::LOG_DAEMON);

sub log_warnings {
	my $fmt = $class eq 'Xyzzy' ? '%s' : "($class) \%s";
	if(-t STDERR) {
		STDERR->autoflush(0);
		foreach my $line (split("\n", $_[0])) {
			$line =~ s/\s+$//;
			next unless $line;
			print STDERR '['.localtime()."] [$class] $line\n";
			$line =~ s/\t/    /g;
			syslog(Sys::Syslog::LOG_WARNING, $fmt, $line);
		}
		STDERR->flush;
		STDERR->autoflush(1);
	} else {
		foreach my $line (split("\n", $_[0])) {
			$line =~ s/\s+$//;
			next unless $line;
			$line =~ s/\t/    /g;
			syslog(Sys::Syslog::LOG_WARNING, $fmt, $line);
		}
	}
}

$SIG{__WARN__} = \&log_warnings;

my $benchmark;
my $numreqs;

eval {
	die "Usage: xyzzy <configfile>\n"
		unless $file;

	my $handler = do {
		my $cfg = $class->new;
		$cfg->include($file);
		$class = ref $cfg;
		$benchmark = $cfg->benchmark;
		$cfg->handler
	};

	$0 = "$class (new)";

	if($benchmark) {
		my $cpu_proc_init = clock_gettime(CLOCK_PROCESS_CPUTIME_ID);
		my $wall_proc_init = clock_gettime(CLOCK_MONOTONIC) - $wall_proc_start;

		printf STDERR "initial: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_init, $cpu_proc_init, 100.0 * $cpu_proc_init / $wall_proc_init;
	}

	while(my $cgi = new CGI::Fast) {
		$numreqs++;
		$0 = "$class (serving #$numreqs)";

		my $wall_req_start = clock_gettime(CLOCK_MONOTONIC);
		my $cpu_req_start = clock_gettime(CLOCK_PROCESS_CPUTIME_ID);

		my $oldout = *STDOUT;
		*STDOUT = *STDERR;
		my @output;

		eval {
			my $ctx = new Xyzzy::Context::Root(cfg => $handler, cgi => $cgi);
			my $res = eval { $handler->handle($ctx) };
			if($res) {
				die "not a Xyzzy::Response object\n"
					unless UNIVERSAL::isa($res, 'Xyzzy::Response');
			} else {
				die "no response from handler\n"
					unless $@;
				die $@ unless UNIVERSAL::isa($@, 'Xyzzy::Response');
				$res = $@;
			}

			my $status = $res->status;
			my $content = $res->content;

			unless(defined $content) {
				$res->mimetype('text/html', charset => 'UTF-8');
				$content = $status->html;
			}

			$content = encode_utf8($content)
				if is_utf8($content);

			$res->setheader('Content-Length', length($content));

			push @output, "Status: $status\r\n";
			my $headers = $res->headers;
			while(my ($key, $vals) = each(%$headers)) {
				my $seperate;
				if(@$vals > 1) {
					foreach my $val (@$vals) {
						$seperate = 1
							unless Xyzzy::Header::istoken($val);
					}
				}
				if($seperate) {
					foreach my $val (@$vals) {
						push @output, $key, ': ', $val, "\r\n";
					}
				} else {
					push @output, $key, ': ', join(', ', @$vals), "\r\n";
				}
			}
			push @output, "\r\n", $content;
		};
		if($@) {
			warn $@;
			my $status = new Xyzzy::Status(500);
			@output = $status->response;
		}

		$0 = "$class (output)";

		foreach(@output) {
			syswrite($oldout, $_)
				or die "Writing output: $!\n";
		}
		*STDOUT = $oldout;

		if($benchmark) {
			my $cpu_req_total = clock_gettime(CLOCK_PROCESS_CPUTIME_ID) - $cpu_req_start;
			my $wall_req_total = clock_gettime(CLOCK_MONOTONIC) - $wall_req_start;

			printf STDERR "request: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_req_total, $cpu_req_total, 100.0 * $cpu_req_total / $wall_req_total;
		}

		$0 = "$class (idle #$numreqs)";
	}
};

if($benchmark) {
	my $cpu_proc_total = clock_gettime(CLOCK_PROCESS_CPUTIME_ID);
	my $wall_proc_total = clock_gettime(CLOCK_MONOTONIC) - $wall_proc_start;

	printf STDERR "process: wall=\%.3fs cpu=\%.3fs (\%.2f\%\%)\n", $wall_proc_total, $cpu_proc_total, 100.0 * $cpu_proc_total / $wall_proc_total;
}

if($@) {
	warn $@;
	exit(1);
}
